import ctypes
import struct
import threading
import time

import usb
import usb.backend.libusb1
import usb.core
from scipy.interpolate import PchipInterpolator


class USB485Pro:
    _INDEX_BOARD_TYPE = 0
    _INDEX_DEVICE_ID = 1
    _INDEX_CONTROL_WORD = 2
    _INDEX_OPERATION_MODE = 3
    _INDEX_STATUS_WORD = 4
    _INDEX_TARGET_CURRENT = 5
    _INDEX_ACTUAL_CURRENT = 6
    _INDEX_IO_OUT = 23
    _INDEX_KPP = 17
    _INDEX_KPI = 18
    _INDEX_KVF = 19
    _INDEX_KFF = 20
    _INDEX_PHASE_CORRECT_CURRENT = 21
    _INDEX_HOMING_TRIGGER = 13
    _INDEX_HOMING_DIRECTION = 14
    _INDEX_HOMING_LEVEL = 15
    _INDEX_ACC_TIME = 11
    _INDEX_SYNC_INTERPOLATION_TARGET_POSITION = 12
    _INDEX_TARGET_VELOCITY = 7
    _INDEX_TARGET_POSITION = 9
    _INDEX_ACTUAL_VELOCITY = 8
    _INDEX_ACTUAL_POSITION = 10
    _INDEX_IO_INPUT = 22
    _INDEX_ENCODER_OFFSET = 24
    _INDEX_ENCODER_POLARITY = 25
    _INDEX_ENCODER_VALUE = 26
    _INDEX_ABS_ENCODER_VALUE = 27
    _INDEX_CURRENT_MAX = 28
    _INDEX_POSITION_ERROR_MAX = 29
    _INDEX_HOMING_OFFSET = 16

    _INDEX_RUNNING_CURRENT = 17
    _INDEX_KEEPING_CURRENT = 18

    _INDEX_CURRENT_BASE = 17
    _INDEX_CURRENT_P = 18
    _INDEX_CURRENT_N = 19
    _INDEX_POWER_LIMIT = 28
    _INDEX_LED_OPTION = 29
    _INDEX_SYSTEM_COUNTER = 30
    _INDEX_LED_COUNTER = 31

    _FUNC_WRITE = 1
    _FUNC_READ = 0
    _FUNC_WRITE_OK = 3
    _FUNC_READ_OK = 2
    _FUNC_OPERATION = 4
    _FUNC_OPERATION_OK = 5
    _FUNC_CHECK = 254
    _FUNC_FREE = 255

    _OPERATION_MODE_PROFILE_VELOCITY = 21
    _OPERATION_MODE_PROFILE_POSITION = 31
    _OPERATION_MODE_HOMING = 40
    _OPERATION_SYNC_INTERPOLATION_POSITION = 34
    _OPERATION_MODE_INTERPOLATION_POSITION = 34
    _OPERATION_MODE_OPEN_LOOP = 35
    _OPERATION_INDEX_MEMORY = 1
    _OPERATION_INDEX_TUNING = 2
    _STATUS_DEVICE_ENABLE = 0X01
    _STATUS_HOMG_FIND = 0X02
    _STATUS_TARGET_REACHED = 0X04
    _STATUS_IO_INPUT = 0X08
    _STATUS_ERROR_OVERCURRENT = 0X20
    _STATUS_ERROR_OVERPOSITION = 0X40
    _BOARD_TYPE_STEPPER_ANT = 0X10
    _BOARD_TYPE_STEPPER_BEE = 0X11
    _BOARD_TYPE_STEPPER_ELEPHANT = 0X12
    _BOARD_TYPE_BDCS_BEE = 0X13
    _BOARD_TYPE_BDC_BEE = 0X14
    _BOARD_TYPE_BLDCS_BEE = 0X15

    def __init__(self):
        self._array_p = []
        for i in range(0, 8):
            self._array_p.append([])
        self.buf_ready_flag = 0

        self._current_pos = [0, 0, 0, 0, 0, 0, 0, 0]
        self._last_si_pos = [0, 0, 0, 0, 0, 0, 0, 0]
        self.msg32 = bytearray(32)

        self._tx_queue = []
        self._tx_message = bytearray(8)
        self._array = (((ctypes.c_int32 * 32) * 8) * 64)
        self._motors = self._array()
        self.dev = usb.core.find(idVendor=0x1a86, idProduct=0x5425)
        self._thread_stop_flag = 0
        self._thread1 = threading.Thread(target=self._linkProcess)
        self._thread1.start()

    def _delay(self):
        time.sleep(0.05)

    # Send Message Function, Users would Call this Function to Send Messages
    def _sendMessage(self, func_code, index, id, subid, data):
        message = bytearray(8)
        struct.pack_into('B', message, 0, *(func_code,))
        struct.pack_into('B', message, 1, *(index,))
        struct.pack_into('B', message, 2, *(id,))
        struct.pack_into('B', message, 3, *(subid,))
        struct.pack_into('i', message, 4, int(*(data,)))
        self._tx_queue.append(message)

    def _analysis(self, rx_message):
        if len(rx_message) == 8:
            func_code = struct.unpack_from('B', rx_message, 0)
            index = struct.unpack_from('B', rx_message, 1)
            id = struct.unpack_from('B', rx_message, 2)
            subid = struct.unpack_from('B', rx_message, 3)
            data = struct.unpack_from('i', rx_message, 4)

            if func_code[0] == self._FUNC_READ_OK:
                if id[0] < 64:
                    if subid[0] < 8:
                        if index[0] < 32:
                            self._motors[id[0]][subid[0]][index[0]] = data[0]

    def setPowerOff(self, id):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_CONTROL_WORD, id, 0, 0)

    def setPowerOn(self, id):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_CONTROL_WORD, id, 0, 1)

    def setRunningCurrent(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_RUNNING_CURRENT, id, 0, value)

    def setKeepingCurrent(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_KEEPING_CURRENT, id, 0, value)

    def setHomingLevel(self, id, value):
        if value == 1:
            self._sendMessage(self._FUNC_WRITE, self._INDEX_HOMING_LEVEL, id, 0, 1)
        elif value == 0:
            self._sendMessage(self._FUNC_WRITE, self._INDEX_HOMING_LEVEL, id, 0, 0)
        else:
            print("wrong value, please try 1 or 0.")

    def setHomingTriggerLevel(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_HOMING_TRIGGER, id, 0, value)

    def setHomingDirection(self, id, value):
        if value == 1:
            self._sendMessage(self._FUNC_WRITE, self._INDEX_HOMING_DIRECTION, id, 0, 1)
        elif value == -1:
            self._sendMessage(self._FUNC_WRITE, self._INDEX_HOMING_DIRECTION, id, 0, -1)
        else:
            print("wrong value, please try 1 or -1.")

    def setHomingMode(self, id):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_OPERATION_MODE, id, 0, self._OPERATION_MODE_HOMING)

    def setHomingModeBroadcast(self):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_OPERATION_MODE, 0XFE, 0, self._OPERATION_MODE_HOMING)

    def setTargetVelocityBroadcast(self, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_TARGET_VELOCITY, 0xfe, 0, value)

    def setHomingOffset(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_HOMING_OFFSET, id, 0, value)

    def setOpenLoopMode(self, id):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_OPERATION_MODE, id, 0, self._OPERATION_MODE_OPEN_LOOP)

    def setAccTime(self, id, value):
        self._motors[id][0][self._INDEX_ACC_TIME] = value
        self._sendMessage(self._FUNC_WRITE, self._INDEX_ACC_TIME, id, 0, value)

    def setOutputIO(self, id, value):
        self._motors[id][0][self._INDEX_IO_OUT] = value
        self._sendMessage(self._FUNC_WRITE, self._INDEX_IO_OUT, id, 0, value)

    def setTargetVelocity(self, id, value):
        # Note: Unit: pulse/ms (51200 pulse per round), this unit nearly equals to RPM
        #       and for stepper motors, 0 to 3000 is reasonable, higher speed will lose steps
        self._motors[id][0][self._INDEX_TARGET_VELOCITY] = value
        self._sendMessage(self._FUNC_WRITE, self._INDEX_TARGET_VELOCITY, id, 0, value)

    def setTargetPosition(self, id, value):
        # Note: Unit: pulse/ms (51200 pulse per round), this unit nearly equals to RPM
        #       and for stepper motors, 0 to 3000 is reasonable, higher speed will lose steps
        self._motors[id][0][self._INDEX_TARGET_POSITION] = value
        self._sendMessage(self._FUNC_WRITE, self._INDEX_TARGET_POSITION, id, 0, value)

    def setEncoderPolarityP(self, id):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_ENCODER_POLARITY, id, 0, 1)

    def setEncoderPolarityN(self, id):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_ENCODER_POLARITY, id, 0, -1)

    def setPhaseCorrectCurrent(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_PHASE_CORRECT_CURRENT, id, 0, value)

    def setKPP(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_KPP, id, 0, value)

    def setKPI(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_KPI, id, 0, value)

    def setKVF(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_KVF, id, 0, value)

    def setKFF(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_KFF, id, 0, value)

    def tunePhase(self, id):
        self._sendMessage(self._FUNC_OPERATION, self._OPERATION_INDEX_TUNING, id, 0, 1)

    def setCurrentBase(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_CURRENT_BASE, id, 0, value)

    def setCurrentP(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_CURRENT_P, id, 0, value)

    def setCurrentN(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_CURRENT_N, id, 0, value)

    def setInterpolationPositionMode(self, id):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_OPERATION_MODE, id, 0,
                          self._OPERATION_MODE_INTERPOLATION_POSITION)

    def setAccTime(self, id, value):
        # Note: acc time is a parameter for accelation and deaccelation progress, unit is ms, normally 200ms to 1000ms is reasonable
        self._sendMessage(self._FUNC_WRITE, self._INDEX_ACC_TIME, id, 0, value)

    def setPositionMode(self, id):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_OPERATION_MODE, id, 0,
                          self._OPERATION_MODE_PROFILE_POSITION)

    def saveParameters(self, id):
        self._sendMessage(self._FUNC_OPERATION, self._OPERATION_INDEX_MEMORY, id, 0, 1)

    def changeID(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_DEVICE_ID, id, 0, value)
        self._delay()
        self.saveParameters(value)

    def getActualVelocity(self, id):
        self._sendMessage(self._FUNC_READ, self._INDEX_ACTUAL_VELOCITY, id, 0, 0)
        self._delay()
        return self._motors[id][0][self._INDEX_ACTUAL_VELOCITY]

    def getActualPosition(self, id):
        self._sendMessage(self._FUNC_READ, self._INDEX_ACTUAL_POSITION, id, 0, 0)
        self._delay()
        return self._motors[id][0][self._INDEX_ACTUAL_POSITION]

    def getInputIO(self, id):
        self._sendMessage(self._FUNC_READ, self._INDEX_IO_INPUT, id, 0, 0)
        self._delay()
        return self._motors[id][0][self._INDEX_IO_INPUT]

    def waitTargetPositionReached(self, id):
        condition = 1
        while condition:
            vel = self.getActualVelocity(id)
            if vel == 0:
                condition = 0

    def getAbsEncoderValue(self, mid):
        self._sendMessage(self._FUNC_READ, self._INDEX_ABS_ENCODER_VALUE, mid, 0, 0)
        self._delay()
        return self._motors[mid][0][self._INDEX_ABS_ENCODER_VALUE]

    def getEncoderValue(self, mid):
        self._sendMessage(self._FUNC_READ, self._INDEX_ENCODER_VALUE, mid, 0, 0)
        self._delay()
        return self._motors[mid][0][self._INDEX_ENCODER_VALUE]

    def getDeviceID(self, id):
        self._sendMessage(self._FUNC_READ, self._INDEX_DEVICE_ID, id, 0, 0)
        self._delay()
        return self._motors[id][0][self._INDEX_DEVICE_ID]

    def getOPMode(self, id):
        self._motors[id][0][self._INDEX_OPERATION_MODE] = 0
        self._sendMessage(self._FUNC_READ, self._INDEX_OPERATION_MODE, id, 0, 0)
        self._delay()
        return self._motors[id][0][self._INDEX_OPERATION_MODE]

    def setPowerLimit(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_POWER_LIMIT, id, 0, value)

    def setLedOption(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_LED_OPTION, id, 0, value)

    def setLedCounter(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_LED_COUNTER, id, 0, value)

    def setSystemCounter(self, id, value):
        self._sendMessage(self._FUNC_WRITE, self._INDEX_SYSTEM_COUNTER, id, 0, value)

    def syncActualPosition(self):
        for mid in range(0, 8):
            self._last_si_pos[mid] = self.getActualPosition(mid)

    def scanDevices(self):
        online = []
        print('Searching Online Devices...')
        for i in range(0, 64):
            self._motors[i][0][self._INDEX_OPERATION_MODE] = 0
        for i in range(0, 64):
            if self.getOPMode(i) != 0:
                online.append(i)
        print(len(online), 'Online Devices:', online)
        return online

    def _pop(self):
        for axis in range(0, 8):
            if len(self._array_p[axis]) > 0:
                struct.pack_into('i', self.msg32, 4 * axis, int(*(self._array_p[axis].pop(0),)))

    def _linkProcess(self):
        while self._thread_stop_flag == 0:
            if len(self._tx_queue) > 0:
                msg = self._tx_queue.pop(0)
                # print(msg)
                # try:
                self.dev.write(0x1, msg, 100)
                # time.sleep(0.005)
                rx = self.dev.read(0x82, 8, 100)
                # except:
                #     print("usb rtx error")
                # print(rx)
                self._analysis(rx)
            else:
                if self.buf_ready_flag == 1:
                    if len(self._array_p[0]) > 0:
                        # time_start = time.time()
                        self._pop()
                        self.dev.write(0x1, self.msg32, 0)
                        rx = self.dev.read(0x82, 32, 0)
                        # time_end = time.time()

                        # print(rx[2])
                        if rx[2] > 6:
                            time.sleep(0.02)
                    # else:
                    # time_c = time_end - time_start
                    # print('time cost', time_c*1000, 'ms')
                    #     p0 =
                    # print(rx[2], len(self._array_p_0), p0)

                else:
                    time.sleep(0.01)

    # wait SI Position Reached
    def waitSIP(self):
        while len(self._array_p[0]) != 0:
            # print(len(self._array_p[0]))
            time.sleep(0.01)

    # set sync interpolation position
    def setSIPose(self, dt, pos):
        self.buf_ready_flag = 0
        for axis in range(0, 8):
            pos_array = []
            tl = []
            tl.append(0)
            pos_array.append(self._last_si_pos[axis])
            tl.append(1)
            pos_array.append(self._last_si_pos[axis])
            tl.append(dt - 1)
            pos_array.append(pos[axis])
            tl.append(dt)
            pos_array.append(pos[axis])
            self._last_si_pos[axis] = pos[axis]
            cs = PchipInterpolator(tl, pos_array)
            x = range(0, dt, 1)
            y = cs(x)
            # print(axis, y.shape[0])
            for i in range(0, y.shape[0]):
                self._array_p[axis].append(y[i])
        self.buf_ready_flag = 1

    # def setSIBuffer(self, trajectory):
    #     self.buf_ready_flag = 0
    #     for axis in range(0, 8):
    #         for i in range(0, trajectory.shape[0]):
    #             self._array_p[axis].append(trajectory[i])
    #     self.buf_ready_flag = 1

    def stop(self):
        time.sleep(1)
        self._thread_stop_flag = 1
